"""
Create kmz products based on SICD type reader.

For a basic help on the command-line, check

>>> python -m sarpy.utils.create_kmz --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import argparse
import logging
import os

import sarpy
from sarpy.io.complex.converter import open_complex
from sarpy.io.product.kmz_product_creation import create_kmz_view


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description="Create derived product is SIDD format from a SICD type file.",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file', metavar='input_file',
        help='Path input data file, or directory for radarsat, RCM, or sentinel.\n'
             '* For radarsat or RCM, this can be the product.xml file, or parent directory\n'
             '  of product.xml or metadata/product.xml.\n'
             '* For sentinel, this can be the manifest.safe file, or parent directory of\n'
             '  manifest.safe.\n')
    parser.add_argument(
        'output_directory', metavar='output_directory',
        help='Path to the output directory where the product file(s) will be created.\n'
             'This directory MUST exist.\n'
             '* Depending on the input file, multiple product files may be produced.\n'
             '* The name for the ouput file(s) will be chosen based on CoreName and\n '
             '  transmit/collect polarization.\n')
    parser.add_argument(
        '-s', '--size', default=3072, type=int, help='Maximum size for the interpolated image, put -1 for full size')
    parser.add_argument(
        '-v', '--verbose', action='store_true', help='Verbose (level="INFO") logging?')

    args = parser.parse_args()

    level = 'INFO' if args.verbose else 'WARNING'
    logging.basicConfig(level=level)
    logger = logging.getLogger('sarpy')
    logger.setLevel(level)

    reader = open_complex(args.input_file)
    file_stem = os.path.splitext(os.path.split(args.input_file)[1])[0]
    pixel_limit = None if args.size == -1 else args.size
    create_kmz_view(reader, args.output_directory, pixel_limit=pixel_limit, file_stem='View-{}'.format(file_stem))



"""
Base structures for phase history readers and usage
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, Tuple

import numpy

from sarpy.io.general.base import AbstractReader
from sarpy.io.phase_history.cphd1_elements.CPHD import CPHDType as CPHDType1_0
from sarpy.io.phase_history.cphd0_3_elements.CPHD import CPHDType as CPHDType0_3


class CPHDTypeReader(AbstractReader):
    """
    An abstract class for ensuring common CPHD functionality.

    This is intended to be used solely in conjunction with implementing a
    legitimate reader.
    """

    def __init__(self, cphd_meta):
        """

        Parameters
        ----------
        cphd_meta : None|CPHDType1_0|CPHDType0_3
            `None`, the CPHD metadata object
        """

        if cphd_meta is None:
            self._cphd_meta = None
        elif isinstance(cphd_meta, (CPHDType1_0, CPHDType0_3)):
            self._cphd_meta = cphd_meta
        else:
            raise TypeError(
                'The cphd_meta must be of type CPHDType, got `{}`'.format(type(cphd_meta)))

    @property
    def cphd_meta(self):
        # type: () -> Union[None, CPHDType1_0, CPHDType0_3]
        """
        None|CPHDType1_0|CPHDType0_3: the cphd meta_data.
        """

        return self._cphd_meta

    def read_support_array(self, index, dim1_range, dim2_range):
        # type: (Union[int, str], Union[None, int, Tuple[int, int], Tuple[int, int, int]], Union[None, int, Tuple[int, int], Tuple[int, int, int]]) -> numpy.ndarray
        """
        Read the support array.

        Parameters
        ----------
        index : int|str
            The support array integer index.
        dim1_range : None|int|Tuple[int, int]|Tuple[int, int, int]
            The row data selection of the form `[start, [stop, [stride]]]`, and
            `None` defaults to all rows (i.e. `(0, NumRows, 1)`)
        dim2_range : None|int|Tuple[int, int]|Tuple[int, int, int]
            The column data selection of the form `[start, [stop, [stride]]]`, and
            `None` defaults to all rows (i.e. `(0, NumCols, 1)`)

        Returns
        -------
        numpy.ndarray

        Raises
        ------
        TypeError
            If called on a reader which doesn't support this.
        """

        raise TypeError('Class {} does not provide support arrays'.format(type(self)))

    def read_support_block(self):
        """
        Reads the entirety of support block(s).

        Returns
        -------
        Dict[str, numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        raise TypeError('Class {} does not provide support arrays'.format(type(self)))

    def read_pvp_variable(self, variable, index, the_range=None):
        """
        Read the vector parameter for the given `variable` and CPHD channel.

        Parameters
        ----------
        variable : str
        index : int|str
            The channel index or identifier.
        the_range : None|int|List[int]|Tuple[int]
            The indices for the vector parameter. `None` returns all, otherwise
            a slice in the (non-traditional) form `([start, [stop, [stride]]])`.

        Returns
        -------
        None|numpy.ndarray
            This will return None if there is no such variable, otherwise the data.
        """

        raise NotImplementedError

    def read_pvp_array(self, index, the_range=None):
        """
        Read the PVP array from the requested channel.

        Parameters
        ----------
        index : int|str
            The support array integer index (of cphd.Data.Channels list) or identifier.
        the_range : None|int|List[int]|Tuple[int]
            The indices for the vector parameter. `None` returns all, otherwise
            a slice in the (non-traditional) form `([start, [stop, [stride]]])`.

        Returns
        -------
        pvp_array : numpy.ndarray
        """

        raise NotImplementedError

    def read_pvp_block(self):
        """
        Reads the entirety of the PVP block(s).

        Returns
        -------
        Dict[Union[int, str], numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the PVP arrays.
        """

        raise NotImplementedError

    def read_signal_block(self):
        """
        Reads the entirety of signal block(s).

        Returns
        -------
        Dict[Union[int, str], numpy.ndarray]
            Dictionary of `numpy.ndarray` containing the support arrays.
        """

        raise NotImplementedError



"""
The TimelineType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import List, Union

import numpy

from sarpy.io.xml.base import Serializable, SerializableArray
from sarpy.io.xml.descriptors import FloatDescriptor, IntegerDescriptor, \
    DateTimeDescriptor, SerializableDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import Poly1DType


class IPPSetType(Serializable):
    """
    The Inter-Pulse Parameter array element container.
    """

    # NOTE that this is simply defined as a child class ("Set") of the TimelineType in the SICD standard
    #   Defining it at root level clarifies the documentation, and giving it a more descriptive name is
    #   appropriate.
    _fields = ('TStart', 'TEnd', 'IPPStart', 'IPPEnd', 'IPPPoly', 'index')
    _required = _fields
    _set_as_attribute = ('index', )
    _numeric_format = {'TStart': FLOAT_FORMAT, 'TEnd': FLOAT_FORMAT, }
    # descriptors
    TStart = FloatDescriptor(
        'TStart', _required, strict=DEFAULT_STRICT,
        docstring='IPP start time relative to collection start time, i.e. offsets in seconds.')  # type: float
    TEnd = FloatDescriptor(
        'TEnd', _required, strict=DEFAULT_STRICT,
        docstring='IPP end time relative to collection start time, i.e. offsets in seconds.')  # type: float
    IPPStart = IntegerDescriptor(
        'IPPStart', _required, strict=True, docstring='Starting IPP index for the period described.')  # type: int
    IPPEnd = IntegerDescriptor(
        'IPPEnd', _required, strict=True, docstring='Ending IPP index for the period described.')  # type: int
    IPPPoly = SerializableDescriptor(
        'IPPPoly', Poly1DType, _required, strict=DEFAULT_STRICT,
        docstring='IPP index polynomial coefficients yield IPP index as a function of time.')  # type: Poly1DType
    index = IntegerDescriptor(
        'index', _required, strict=DEFAULT_STRICT, docstring='The element array index.')  # type: int

    def __init__(self, TStart=None, TEnd=None, IPPStart=None, IPPEnd=None, IPPPoly=None, index=None, **kwargs):
        """

        Parameters
        ----------
        TStart : float
        TEnd : float
        IPPStart : int
        IPPEnd : int
        IPPPoly : Poly1DType|numpy.ndarray|list|tuple
        index : int
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TStart, self.TEnd = TStart, TEnd
        self.IPPStart, self.IPPEnd = IPPStart, IPPEnd
        self.IPPPoly = IPPPoly
        self.index = index
        super(IPPSetType, self).__init__(**kwargs)

    def _basic_validity_check(self):
        condition = super(IPPSetType, self)._basic_validity_check()
        if self.TStart >= self.TEnd:
            self.log_validity_error(
                'TStart ({}) >= TEnd ({})'.format(self.TStart, self.TEnd))
            condition = False

        if self.IPPStart >= self.IPPEnd:
            self.log_validity_error(
                'IPPStart ({}) >= IPPEnd ({})'.format(self.IPPStart, self.IPPEnd))
            condition = False

        exp_ipp_start = self.IPPPoly(self.TStart)
        exp_ipp_end = self.IPPPoly(self.TEnd)
        if abs(exp_ipp_start - self.IPPStart) > 1:
            self.log_validity_error(
                'IPPStart populated as {}, inconsistent with value ({}) '
                'derived from IPPPoly and TStart'.format(exp_ipp_start, self.IPPStart))
        if abs(exp_ipp_end - self.IPPEnd) > 1:
            self.log_validity_error(
                'IPPEnd populated as {}, inconsistent with value ({}) '
                'derived from IPPPoly and TEnd'.format(self.IPPEnd, exp_ipp_end))
        return condition


class TimelineType(Serializable):
    """
    The details for the imaging collection timeline.
    """

    _fields = ('CollectStart', 'CollectDuration', 'IPP')
    _required = ('CollectStart', 'CollectDuration', )
    _collections_tags = {'IPP': {'array': True, 'child_tag': 'Set'}}
    _numeric_format = {'CollectDuration': FLOAT_FORMAT, }
    # descriptors
    CollectStart = DateTimeDescriptor(
        'CollectStart', _required, strict=DEFAULT_STRICT, numpy_datetime_units='us',
        docstring='The collection start time. The default precision will be microseconds.')  # type: numpy.datetime64
    CollectDuration = FloatDescriptor(
        'CollectDuration', _required, strict=DEFAULT_STRICT,
        docstring='The duration of the collection in seconds.')  # type: float
    IPP = SerializableArrayDescriptor(
        'IPP', IPPSetType, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=1,
        docstring="The Inter-Pulse Period (IPP) parameters array.")  # type: Union[SerializableArray, List[IPPSetType]]

    def __init__(self, CollectStart=None, CollectDuration=None, IPP=None, **kwargs):
        """

        Parameters
        ----------
        CollectStart : numpy.datetime64|datetime|date|str
        CollectDuration : float
        IPP : List[IPPSetType]
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CollectStart = CollectStart
        self.CollectDuration = CollectDuration
        self.IPP = IPP
        super(TimelineType, self).__init__(**kwargs)

    @property
    def CollectEnd(self):
        """
        None|numpy.datetime64: The collection end time, inferred from `CollectEnd` and `CollectDuration`,
        provided that both are populated.
        """

        if self.CollectStart is None or self.CollectDuration is None:
            return None

        return self.CollectStart + numpy.timedelta64(int(self.CollectDuration*1e6), 'us')

    def _check_ipp_consecutive(self):
        if self.IPP is None or len(self.IPP) < 2:
            return True
        cond = True
        for i in range(len(self.IPP)-1):
            el1 = self.IPP[i]
            el2 = self.IPP[i+1]
            if el1.IPPEnd + 1 != el2.IPPStart:
                self.log_validity_error(
                    'IPP entry {} IPPEnd ({}) is not consecutive with '
                    'entry {} IPPStart ({})'.format(i, el1.IPPEnd, i+1, el2.IPPStart))
                cond = False
            if el1.TEnd >= el2.TStart:
                self.log_validity_error(
                    'IPP entry {} TEnd ({}) is greater than entry {} TStart ({})'.format(i, el1.TEnd, i+1, el2.TStart))
        return cond

    def _check_ipp_times(self):
        if self.IPP is None:
            return True

        cond = True
        min_time = self.IPP[0].TStart
        max_time = self.IPP[0].TEnd
        for i in range(len(self.IPP)):
            entry = self.IPP[i]
            if entry.TStart < 0:
                self.log_validity_error('IPP entry {} has negative TStart ({})'.format(i, entry.TStart))
                cond = False
            if entry.TEnd > self.CollectDuration + 1e-2:
                self.log_validity_error(
                    'IPP entry {} has TEnd ({}) appreciably larger than '
                    'CollectDuration ({})'.format(i, entry.TEnd, self.CollectDuration))
                cond = False
            min_time = min(min_time, entry.TStart)
            max_time = max(max_time, entry.TEnd)
        if abs(max_time - min_time - self.CollectDuration) > 1e-2:
            self.log_validity_error(
                'time range in IPP entries ({}) not in keeping with populated '
                'CollectDuration ({})'.format(max_time-min_time, self.CollectDuration))
            cond = False
        return cond

    def _basic_validity_check(self):
        condition = super(TimelineType, self)._basic_validity_check()
        condition &= self._check_ipp_consecutive()
        condition &= self._check_ipp_times()
        return condition



"""
The ErrorStatisticsType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from sarpy.io.xml.base import Serializable, ParametersCollection
from sarpy.io.xml.descriptors import StringEnumDescriptor, FloatDescriptor, \
    SerializableDescriptor, ParametersDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import ErrorDecorrFuncType


class CompositeSCPErrorType(Serializable):
    """
    Composite error statistics for the Scene Center Point. Slant plane range *(Rg)*
    and azimuth *(Az)* error statistics. Slant plane defined at *SCP COA*.
    """
    _fields = ('Rg', 'Az', 'RgAz')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # descriptors
    Rg = FloatDescriptor(
        'Rg', _required, strict=DEFAULT_STRICT,
        docstring='Estimated range error standard deviation.')  # type: float
    Az = FloatDescriptor(
        'Az', _required, strict=DEFAULT_STRICT,
        docstring='Estimated azimuth error standard deviation.')  # type: float
    RgAz = FloatDescriptor(
        'RgAz', _required, strict=DEFAULT_STRICT,
        docstring='Estimated range and azimuth error correlation coefficient.')  # type: float

    def __init__(self, Rg=None, Az=None, RgAz=None, **kwargs):
        """

        Parameters
        ----------
        Rg : float
        Az : float
        RgAz : float
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Rg, self.Az, self.RgAz = Rg, Az, RgAz
        super(CompositeSCPErrorType, self).__init__(**kwargs)


class CorrCoefsType(Serializable):
    """Correlation Coefficient parameters."""
    _fields = (
        'P1P2', 'P1P3', 'P1V1', 'P1V2', 'P1V3', 'P2P3', 'P2V1', 'P2V2', 'P2V3',
        'P3V1', 'P3V2', 'P3V3', 'V1V2', 'V1V3', 'V2V3')
    _required = _fields
    _numeric_format = {key: FLOAT_FORMAT for key in _fields}
    # descriptors
    P1P2 = FloatDescriptor(
        'P1P2', _required, strict=DEFAULT_STRICT, docstring='`P1` and `P2` correlation coefficient.')  # type: float
    P1P3 = FloatDescriptor(
        'P1P3', _required, strict=DEFAULT_STRICT, docstring='`P1` and `P3` correlation coefficient.')  # type: float
    P1V1 = FloatDescriptor(
        'P1V1', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V1` correlation coefficient.')  # type: float
    P1V2 = FloatDescriptor(
        'P1V2', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V2` correlation coefficient.')  # type: float
    P1V3 = FloatDescriptor(
        'P1V3', _required, strict=DEFAULT_STRICT, docstring='`P1` and `V3` correlation coefficient.')  # type: float
    P2P3 = FloatDescriptor(
        'P2P3', _required, strict=DEFAULT_STRICT, docstring='`P2` and `P3` correlation coefficient.')  # type: float
    P2V1 = FloatDescriptor(
        'P2V1', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V1` correlation coefficient.')  # type: float
    P2V2 = FloatDescriptor(
        'P2V2', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V2` correlation coefficient.')  # type: float
    P2V3 = FloatDescriptor(
        'P2V3', _required, strict=DEFAULT_STRICT, docstring='`P2` and `V3` correlation coefficient.')  # type: float
    P3V1 = FloatDescriptor(
        'P3V1', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V1` correlation coefficient.')  # type: float
    P3V2 = FloatDescriptor(
        'P3V2', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V2` correlation coefficient.')  # type: float
    P3V3 = FloatDescriptor(
        'P3V3', _required, strict=DEFAULT_STRICT, docstring='`P3` and `V3` correlation coefficient.')  # type: float
    V1V2 = FloatDescriptor(
        'V1V2', _required, strict=DEFAULT_STRICT, docstring='`V1` and `V2` correlation coefficient.')  # type: float
    V1V3 = FloatDescriptor(
        'V1V3', _required, strict=DEFAULT_STRICT, docstring='`V1` and `V3` correlation coefficient.')  # type: float
    V2V3 = FloatDescriptor(
        'V2V3', _required, strict=DEFAULT_STRICT, docstring='`V2` and `V3` correlation coefficient.')  # type: float

    def __init__(self, P1P2=None, P1P3=None, P1V1=None, P1V2=None, P1V3=None,
                 P2P3=None, P2V1=None, P2V2=None, P2V3=None,
                 P3V1=None, P3V2=None, P3V3=None,
                 V1V2=None, V1V3=None,
                 V2V3=None, **kwargs):
        """

        Parameters
        ----------
        P1P2 : float
        P1P3 : float
        P1V1 : float
        P1V2 : float
        P1V3 : float
        P2P3 : float
        P2V1 : float
        P2V2 : float
        P2V3 : float
        P3V1 : float
        P3V2 : float
        P3V3 : float
        V1V2 : float
        V1V3 : float
        V2V3 : float
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.P1P2, self.P1P3, self.P1V1, self.P1V2, self.P1V3 = P1P2, P1P3, P1V1, P1V2, P1V3
        self.P2P3, self.P2V1, self.P2V2, self.P2V3 = P2P3, P2V1, P2V2, P2V3
        self.P3V1, self.P3V2, self.P3V3 = P3V1, P3V2, P3V3
        self.V1V2, self.V1V3 = V1V2, V1V3
        self.V2V3 = V2V3
        super(CorrCoefsType, self).__init__(**kwargs)


class PosVelErrType(Serializable):
    """
    Position and velocity error statistics for the radar platform.
    """

    _fields = ('Frame', 'P1', 'P2', 'P3', 'V1', 'V2', 'V3', 'CorrCoefs', 'PositionDecorr')
    _required = ('Frame', 'P1', 'P2', 'P3', 'V1', 'V2', 'V3')
    _numeric_format = {
        'P1': FLOAT_FORMAT, 'P2': FLOAT_FORMAT, 'P3': FLOAT_FORMAT,
        'V1': FLOAT_FORMAT, 'V2': FLOAT_FORMAT, 'V3': FLOAT_FORMAT}
    # class variables
    _FRAME_VALUES = ('ECF', 'RIC_ECF', 'RIC_ECI')
    # descriptors
    Frame = StringEnumDescriptor(
        'Frame', _FRAME_VALUES, _required, strict=True,
        docstring='Coordinate frame used for expressing P,V errors statistics. Note: '
                  '*RIC = Radial, In-Track, Cross-Track*, where radial is defined to be from earth center through '
                  'the platform position.')  # type: str
    P1 = FloatDescriptor(
        'P1', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 1 standard deviation.')  # type: float
    P2 = FloatDescriptor(
        'P2', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 2 standard deviation.')  # type: float
    P3 = FloatDescriptor(
        'P3', _required, strict=DEFAULT_STRICT, docstring='Position coordinate 3 standard deviation.')  # type: float
    V1 = FloatDescriptor(
        'V1', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 1 standard deviation.')  # type: float
    V2 = FloatDescriptor(
        'V2', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 2 standard deviation.')  # type: float
    V3 = FloatDescriptor(
        'V3', _required, strict=DEFAULT_STRICT, docstring='Velocity coordinate 3 standard deviation.')  # type: float
    CorrCoefs = SerializableDescriptor(
        'CorrCoefs', CorrCoefsType, _required, strict=DEFAULT_STRICT,
        docstring='Correlation Coefficient parameters.')  # type: CorrCoefsType
    PositionDecorr = SerializableDescriptor(
        'PositionDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Platform position error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, Frame=None, P1=None, P2=None, P3=None, V1=None, V2=None, V3=None,
                 CorrCoefs=None, PositionDecorr=None, **kwargs):
        """

        Parameters
        ----------
        Frame : str
        P1 : float
        P2 : float
        P3 : float
        V1 : float
        V2 : float
        V3 : float
        CorrCoefs : CorrCoefsType
        PositionDecorr : ErrorDecorrFuncType
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Frame = Frame
        self.P1, self.P2, self.P3 = P1, P2, P3
        self.V1, self.V2, self.V3 = V1, V2, V3
        self.CorrCoefs, self.PositionDecorr = CorrCoefs, PositionDecorr
        super(PosVelErrType, self).__init__(**kwargs)


class RadarSensorErrorType(Serializable):
    """Radar sensor error statistics."""
    _fields = ('RangeBias', 'ClockFreqSF', 'TransmitFreqSF', 'RangeBiasDecorr')
    _required = ('RangeBias', )
    _numeric_format = {'RangeBias': FLOAT_FORMAT, 'ClockFreqSF': FLOAT_FORMAT, 'TransmitFreqSF': FLOAT_FORMAT}
    # descriptors
    RangeBias = FloatDescriptor(
        'RangeBias', _required, strict=DEFAULT_STRICT,
        docstring='Range bias error standard deviation.')  # type: float
    ClockFreqSF = FloatDescriptor(
        'ClockFreqSF', _required, strict=DEFAULT_STRICT,
        docstring='Payload clock frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    TransmitFreqSF = FloatDescriptor(
        'TransmitFreqSF', _required, strict=DEFAULT_STRICT,
        docstring='Transmit frequency scale factor standard deviation, '
                  r'where :math:`SF = (\Delta f)/f_0`.')  # type: float
    RangeBiasDecorr = SerializableDescriptor(
        'RangeBiasDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Range bias decorrelation rate.')  # type: ErrorDecorrFuncType

    def __init__(self, RangeBias=None, ClockFreqSF=None, TransmitFreqSF=None, RangeBiasDecorr=None, **kwargs):
        """

        Parameters
        ----------
        RangeBias : float
        ClockFreqSF : float
        TransmitFreqSF : float
        RangeBiasDecorr : ErrorDecorrFuncType
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RangeBias, self.ClockFreqSF, self.TransmitFreqSF = RangeBias, ClockFreqSF, TransmitFreqSF
        self.RangeBiasDecorr = RangeBiasDecorr
        super(RadarSensorErrorType, self).__init__(**kwargs)


class TropoErrorType(Serializable):
    """Troposphere delay error statistics."""
    _fields = ('TropoRangeVertical', 'TropoRangeSlant', 'TropoRangeDecorr')
    _required = ()
    _numeric_format = {'TropoRangeVertical': FLOAT_FORMAT, 'TropoRangeSlant': FLOAT_FORMAT}
    # descriptors
    TropoRangeVertical = FloatDescriptor(
        'TropoRangeVertical', _required, strict=DEFAULT_STRICT,
        docstring='Troposphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    TropoRangeSlant = FloatDescriptor(
        'TropoRangeSlant', _required, strict=DEFAULT_STRICT,
        docstring='Troposphere two-way delay error for the *SCP* line of sight at *COA* standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    TropoRangeDecorr = SerializableDescriptor(
        'TropoRangeDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere range error decorrelation function.')  # type: ErrorDecorrFuncType

    def __init__(self, TropoRangeVertical=None, TropoRangeSlant=None, TropoRangeDecorr=None, **kwargs):
        """

        Parameters
        ----------
        TropoRangeVertical : float
        TropoRangeSlant : float
        TropoRangeDecorr : ErrorDecorrFuncType
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.TropoRangeVertical, self.TropoRangeSlant = TropoRangeVertical, TropoRangeSlant
        self.TropoRangeDecorr = TropoRangeDecorr
        super(TropoErrorType, self).__init__(**kwargs)


class IonoErrorType(Serializable):
    """Ionosphere delay error statistics."""
    _fields = ('IonoRangeVertical', 'IonoRangeSlant', 'IonoRgRgRateCC', 'IonoRangeDecorr')
    _required = ('IonoRgRgRateCC', )
    _numeric_format = {
        'IonoRangeVertical': FLOAT_FORMAT, 'IonoRangeSlant': FLOAT_FORMAT, 'IonoRgRgRateCC': FLOAT_FORMAT}
    # descriptors
    IonoRangeVertical = FloatDescriptor(
        'IonoRangeVertical', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere two-way delay error for normal incidence standard deviation. '
                  r'Expressed as a range error. :math:`(\Delta R) = (\Delta T) \cdot (c/2)`.')  # type: float
    IonoRangeSlant = FloatDescriptor(
        'IonoRangeSlant', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere two-way delay rate of change error for normal incidence standard deviation. '
                  r'Expressed as a range rate error. '
                  r':math:`(\Delta \dot{R}) = (\Delta \dot{T}) \cdot (c/2)`.')  # type: float
    IonoRgRgRateCC = FloatDescriptor(
        'IonoRgRgRateCC', _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error and range rate error correlation coefficient.')  # type: float
    IonoRangeDecorr = SerializableDescriptor(
        'IonoRangeDecorr', ErrorDecorrFuncType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere range error decorrelation rate.')  # type: ErrorDecorrFuncType

    def __init__(self, IonoRangeVertical=None, IonoRangeSlant=None,
                 IonoRgRgRateCC=None, IonoRangeDecorr=None, **kwargs):
        """

        Parameters
        ----------
        IonoRangeVertical : float
        IonoRangeSlant : float
        IonoRgRgRateCC : float
        IonoRangeDecorr : ErrorDecorrFuncType
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.IonoRangeVertical, self.IonoRangeSlant = IonoRangeVertical, IonoRangeSlant
        self.IonoRgRgRateCC = IonoRgRgRateCC
        self.IonoRangeDecorr = IonoRangeDecorr
        super(IonoErrorType, self).__init__(**kwargs)


class ErrorComponentsType(Serializable):
    """Error statistics by components."""
    _fields = ('PosVelErr', 'RadarSensor', 'TropoError', 'IonoError')
    _required = ('PosVelErr', 'RadarSensor')
    # descriptors
    PosVelErr = SerializableDescriptor(
        'PosVelErr', PosVelErrType, _required, strict=DEFAULT_STRICT,
        docstring='Position and velocity error statistics for the radar platform.')  # type: PosVelErrType
    RadarSensor = SerializableDescriptor(
        'RadarSensor', RadarSensorErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Radar sensor error statistics.')  # type: RadarSensorErrorType
    TropoError = SerializableDescriptor(
        'TropoError', TropoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Troposphere delay error statistics.')  # type: TropoErrorType
    IonoError = SerializableDescriptor(
        'IonoError', IonoErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Ionosphere delay error statistics.')  # type: IonoErrorType

    def __init__(self, PosVelErr=None, RadarSensor=None, TropoError=None, IonoError=None, **kwargs):
        """

        Parameters
        ----------
        PosVelErr : PosVelErrType
        RadarSensor : RadarSensorErrorType
        TropoError : TropoErrorType
        IonoError : IonoErrorType
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PosVelErr, self.RadarSensor = PosVelErr, RadarSensor
        self.TropoError, self.IonoError = TropoError, IonoError
        super(ErrorComponentsType, self).__init__(**kwargs)


class ErrorStatisticsType(Serializable):
    """Parameters used to compute error statistics within the SICD sensor model."""
    _fields = ('CompositeSCP', 'Components', 'AdditionalParms')
    _required = ()
    _collections_tags = {'AdditionalParms': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    CompositeSCP = SerializableDescriptor(
        'CompositeSCP', CompositeSCPErrorType, _required, strict=DEFAULT_STRICT,
        docstring='Composite error statistics for the scene center point. *Slant plane range (Rg)* and *azimuth (Az)* '
                  'error statistics. Slant plane defined at *Scene Center Point, Center of Azimuth (SCP COA)*.'
    )  # type: CompositeSCPErrorType
    Components = SerializableDescriptor(
        'Components', ErrorComponentsType, _required, strict=DEFAULT_STRICT,
        docstring='Error statistics by components.')  # type: ErrorComponentsType

    AdditionalParms = ParametersDescriptor(
        'AdditionalParms', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Any additional parameters.')  # type: ParametersCollection

    def __init__(self, CompositeSCP=None, Components=None, AdditionalParms=None, **kwargs):
        """

        Parameters
        ----------
        CompositeSCP : CompositeSCPErrorType
        Components : ErrorComponentsType
        AdditionalParms : ParametersCollection|dict
        kwargs : dict
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CompositeSCP = CompositeSCP
        self.Components = Components
        self.AdditionalParms = AdditionalParms
        super(ErrorStatisticsType, self).__init__(**kwargs)



"""
Base common features for product readers
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Sequence, List, Tuple, Union

from sarpy.io.general.base import AbstractReader
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.complex.sicd_elements.SICD import SICDType


class SIDDTypeReader(AbstractReader):

    def __init__(self, sidd_meta, sicd_meta):
        """

        Parameters
        ----------
        sidd_meta : None|SIDDType1|SIDDType2|Sequence[SIDDType1]|Sequence[SIDDType2]
            The SIDD metadata object(s), if provided
        sicd_meta : None|SICDType|Sequence[SICDType]
            the SICD metadata object(s), if provided
        """


        if sidd_meta is None:
            self._sidd_meta = None
        elif isinstance(sidd_meta, (SIDDType1, SIDDType2)):
            self._sidd_meta = sidd_meta
        else:
            temp_list = []  # type: List[Union[SIDDType1]]
            for el in sidd_meta:
                if not isinstance(el, (SIDDType1, SIDDType2)):
                    raise TypeError(
                        'Got a collection for sidd_meta, and all elements are required '
                        'to be instances of SIDDType.')
                temp_list.append(el)
            self._sidd_meta = tuple(temp_list)

        if sicd_meta is None:
            self._sicd_meta = None
        elif isinstance(sicd_meta, SICDType):
            self._sicd_meta = (sicd_meta, )
        else:
            temp_list = [] # type: List[SICDType]
            for el in sicd_meta:
                if not isinstance(el, SICDType):
                    raise TypeError(
                        'Got a collection for sicd_meta, and all elements are required '
                        'to be instances of SICDType.')
                temp_list.append(el)
            self._sicd_meta = tuple(temp_list)

    @property
    def sidd_meta(self):
        # type: () -> Union[None, SIDDType1, SIDDType2, Tuple[SIDDType1], Tuple[SIDDType2]]
        """
        None|SIDDType1|SIDDType2|Tuple[SIDDType1]|Tuple[SIDDType2]: the sidd meta_data collection.
        """

        return self._sidd_meta

    @property
    def sicd_meta(self):
        # type: () -> Union[None, Tuple[SICDType]]
        """
        None|Tuple[SICDType]: the sicd meta_data collection.
        """

        return self._sicd_meta

    def get_sidds_as_tuple(self):
        """
        Get the sidd collection as a tuple - for simplicity and consistency of use.

        Returns
        -------
        Tuple[SIDDType1]|Tuple[SIDDType2]
        """

        if self._sidd_meta is None:
            return None
        elif isinstance(self._sidd_meta, tuple):
            return self._sidd_meta
        else:
            return (self._sidd_meta, )



"""
Functionality for reading SIO data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas McCullough", "Wade Schwartzkopf")


import os
import struct
import logging
import re
from typing import Union, Dict, Tuple

import numpy

from sarpy.io.general.base import BaseReader, BIPChipper, BIPWriter, SarpyIOError
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import RowColType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType, FullImageType
from sarpy.io.complex.sicd import complex_to_amp_phase, complex_to_int, amp_phase_to_complex
from sarpy.io.general.utils import is_file_like
from sarpy.io.xml.base import parse_xml_from_string

logger = logging.getLogger(__name__)

_unsupported_pix_size = 'Got unsupported sio data type/pixel size = `{}`'


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name):
    """
    Tests whether a given file_name corresponds to a SIO file. Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    SIOReader|None
        `SIOReader` instance if SIO file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        sio_details = SIODetails(file_name)
        logger.info('File {} is determined to be a SIO file.'.format(file_name))
        return SIOReader(sio_details)
    except SarpyIOError:
        return None


###########
# parser and interpreter for hdf5 attributes

class SIODetails(object):
    __slots__ = (
        '_file_name', '_magic_number', '_head', '_user_data', '_data_offset',
        '_caspr_data', '_symmetry', '_sicd')

    # NB: there are really just two types of SIO file (with user_data and without),
    #   with endian-ness layered on top
    ENDIAN = {
        0xFF017FFE: '>', 0xFE7F01FF: '<',  # no user data
        0xFF027FFD: '>', 0xFD7F02FF: '<'}  # with user data

    def __init__(self, file_name):
        self._file_name = file_name
        self._user_data = None
        self._data_offset = 20
        self._caspr_data = None
        self._symmetry = (False, False, False)
        self._sicd = None

        if not os.path.isfile(file_name):
            raise SarpyIOError('Path {} is not a file'.format(file_name))

        with open(file_name, 'rb') as fi:
            self._magic_number = struct.unpack('I', fi.read(4))[0]
            endian = self.ENDIAN.get(self._magic_number, None)
            if endian is None:
                raise SarpyIOError('File {} is not an SIO file. Got magic number {}'.format(file_name, self._magic_number))

            # reader basic header - (rows, columns, data_type, pixel_size)?
            init_head = numpy.array(struct.unpack('{}4I'.format(endian), fi.read(16)), dtype=numpy.uint64)
            if not (numpy.all(init_head[2:] == numpy.array([13, 8]))
                    or numpy.all(init_head[2:] == numpy.array([12, 4]))
                    or numpy.all(init_head[2:] == numpy.array([11, 2]))):
                raise SarpyIOError(_unsupported_pix_size.format(init_head[2:]))
            self._head = init_head

    @property
    def file_name(self):  # type: () -> str
        return self._file_name

    @property
    def symmetry(self):  # type: () -> Tuple[bool, bool, bool]
        return self._symmetry

    @property
    def data_offset(self):  # type: () -> int
        return self._data_offset

    @property
    def data_size(self):  # type: () -> Union[None, Tuple[int, int]]
        if self._head is None:
            return None
        rows, cols = self._head[:2]
        if self._symmetry[0]:
            return cols, rows
        else:
            return rows, cols

    @property
    def data_type(self):  # type: () -> Union[None, str]
        # head[2] = (2X = vector, 1X = complex/scalar, 0X = real/scalar), where
        #   X = (1 = unsigned int, 2 = signed int, 3 = float, (4=double? I would guess)
        # head[3] = pixel size in bytes (2*bit depth for complex, or band*bit depth for vector)
        pixel_size = self._head[3]
        # we require (for sicd) that either head[2:] is [13, 8], [12, 4], [11, 2]
        if pixel_size == 8:
            return 'float32'
        elif pixel_size == 4:
            return 'int16'
        elif pixel_size == 2:
            return 'uint8'
        else:
            raise ValueError(_unsupported_pix_size.format(self._head[2:]))

    @property
    def pixel_type(self):  # type: () -> Union[None, str]
        if self._head[2] == 13 and self._head[3] == 8:
            return 'RE32F_IM32F'
        elif self._head[2] == 12 and self._head[3] == 4:
            return 'RE16I_IM16I'
        elif self._head[2] == 11 and self._head[3] == 2:
            return 'AMP8I_PHS8I'
        else:
            raise ValueError(_unsupported_pix_size.format(self._head[2:]))

    def _read_user_data(self):
        if self._user_data is not None:
            return
        if self._magic_number in (0xFF017FFE, 0xFE7F01FF):  # no user data
            self._user_data = {}
        else:
            def read_user_data():
                out = {}
                user_dat_len = 0
                if self._magic_number in (0xFF017FFE, 0xFE7F01FF):  # no user data
                    return out, user_dat_len

                num_data_pairs = struct.unpack('{}I'.format(endian), fi.read(4))[0]

                for i in range(num_data_pairs):
                    name_length = struct.unpack('{}I'.format(endian), fi.read(4))[0]
                    name = struct.unpack('{}{}s'.format(endian, name_length), fi.read(name_length))[0].decode('utf-8')
                    value_length = struct.unpack('{}I'.format(endian), fi.read(4))[0]
                    value = struct.unpack('{}{}s'.format(endian, value_length), fi.read(value_length))[0]
                    try:
                        value = value.decode('utf-8')
                    except UnicodeDecodeError:
                        # leave value as bytes - it may just be some other type
                        pass
                    out[name] = value
                    user_dat_len += 4 + name_length + 4 + value_length
                return out, user_dat_len

            endian = self.ENDIAN[self._magic_number]
            with open(self._file_name, 'rb') as fi:
                fi.seek(20, os.SEEK_SET)  # skip the basic header
                # read the user data (some type of header), if necessary
                user_data, user_data_length = read_user_data()
            self._user_data = user_data
            self._data_offset = 20 + user_data_length
        # validate file size
        exp_file_size = self._data_offset + self._head[0]*self._head[1]*self._head[3]
        act_file_size = os.path.getsize(self._file_name)
        if exp_file_size != act_file_size:
            logger.warning(
                'File {} appears to be an SIO file.\n\t'
                'The file size calculated from the header ({})\n\t'
                'does not match the actual file size ({})'.format(
                    self._file_name, exp_file_size, act_file_size))

    def _find_caspr_data(self):
        def find_caspr():
            dir_name, fil_name = os.path.split(self._file_name)
            file_stem = os.path.splitext(fil_name)
            for fil in [os.path.join(dir_name, '{}.hydra'.format(file_stem)),
                        os.path.join(dir_name, '{}.hdr'.format(file_stem)),
                        os.path.join(dir_name, '..', 'RPHDHeader.out'),
                        os.path.join(dir_name, '..', '_RPHDHeader.out')]:
                if os.path.exists(fil) and os.path.isfile(fil):
                    # generally redundant, except for broken link?
                    return fil
            return None

        casp_fil = find_caspr()
        if casp_fil is None:
            return

        out = {}
        with open(casp_fil, 'r') as fi:
            lines = fi.read().splitlines(keepends=False)
        # this is generally just copied from the previous version - maybe refactor eventually
        current_subfield = ''
        reading_subfield = False
        for line in lines:
            if len(line.strip()) == 0:
                continue  # skip blank lines
            if line.startswith(';;;'):  # I guess this is the subfield delimiter?
                reading_subfield = ~reading_subfield  # change state
                if reading_subfield:
                    current_subfield = ''
                else:
                    out[current_subfield] = {}  # prepare the workspace
            else:
                quoted_token = re.match('"(?P<quoted>[^"]+)"', line)
                if quoted_token:  # Some values with spaces are surrounded by quotes
                    tokens = [quoted_token.group('quoted'),
                              line[quoted_token.end('quoted') + 1:].strip()]
                else:  # No quoted values were found
                    # If not using quotes, split with whitespace
                    tokens = line.split(None, 1)
                if (len(tokens) > 1) and tokens[1] != '':
                    if reading_subfield:  # Subsection heading
                        current_subfield = current_subfield + tokens[1]
                    elif not current_subfield == '':  # Actual field value
                        try:
                            out[current_subfield][tokens[1]] = float(tokens[0])
                        except ValueError:  # Value is string not numeric
                            out[current_subfield][tokens[1]] = tokens[0]
        self._caspr_data = out
        # set symmetry
        im_params = out.get('Image Parameters', None)
        if im_params is None:
            return
        illum_dir = im_params.get('image illumination direction [top, left, bottom, right]', None)
        if illum_dir is None:
            return
        elif illum_dir == 'left':
            self._symmetry = (True, False, True)
        elif illum_dir != 'top':
            raise ValueError('unhandled illumination direction {}'.format(illum_dir))

    def get_sicd(self):
        """
        Extract the SICD details.

        Returns
        -------
        SICDType
        """

        if self._sicd is not None:
            return self._sicd
        if self._user_data is None:
            self._read_user_data()
        if self._caspr_data is None:
            self._find_caspr_data()
        # Check if the user data contains a sicd structure.
        sicd_string = None
        for nam in ['SICDMETA', 'SICD_META', 'SICD']:
            if sicd_string is None:
                sicd_string = self._user_data.get(nam, None)
        # If so, assume that this SICD is valid and simply present it
        if sicd_string is not None:
            root_node, xml_ns = parse_xml_from_string(sicd_string)
            self._sicd = SICDType.from_node(root_node, xml_ns)
            self._sicd.derive()
        else:
            # otherwise, we populate a really minimal sicd structure
            num_rows, num_cols = self.data_size
            self._sicd = SICDType(ImageData=ImageDataType(NumRows=num_rows,
                                                          NumCols=num_cols,
                                                          FirstRow=0,
                                                          FirstCol=0,
                                                          PixelType=self.pixel_type,
                                                          FullImage=FullImageType(NumRows=num_rows,
                                                                                  NumCols=num_cols),
                                                          SCPPixel=RowColType(Row=num_rows/2,
                                                                              Col=num_cols/2)))
        return self._sicd


#######
#  The actual reading implementation

class SIOReader(BaseReader, SICDTypeReader):
    __slots__ = ('_sio_details', )

    def __init__(self, sio_details):
        """

        Parameters
        ----------
        sio_details : str|SIODetails
            filename or SIODetails object
        """

        if isinstance(sio_details, str):
            sio_details = SIODetails(sio_details)
        if not isinstance(sio_details, SIODetails):
            raise TypeError('The input argument for SIOReader must be a filename or '
                            'SIODetails object.')
        self._sio_details = sio_details
        sicd_meta = sio_details.get_sicd()
        if sicd_meta.ImageData.PixelType == 'AMP8I_PHS8I':
            transform_data = amp_phase_to_complex(sicd_meta.ImageData.AmpTable)
        else:
            transform_data = 'COMPLEX'
        raw_bands = 2
        output_bands = 1
        output_dtype = 'complex64'
        chipper = BIPChipper(sio_details.file_name, sio_details.data_type, sio_details.data_size,
                             raw_bands, output_bands, output_dtype,
                             symmetry=sio_details.symmetry, transform_data=transform_data,
                             data_offset=sio_details.data_offset)

        SICDTypeReader.__init__(self, sicd_meta)
        BaseReader.__init__(self, chipper, reader_type="SICD")
        self._check_sizes()

    @property
    def sio_details(self):
        # type: () -> SIODetails
        """
        SIODetails: The sio details object.
        """

        return self._sio_details

    @property
    def file_name(self):
        return self.sio_details.file_name


#######
#  The actual writing implementation

class SIOWriter(BIPWriter):
    def __init__(self, file_name, sicd_meta, user_data=None, check_older_version=False, check_existence=True):
        """

        Parameters
        ----------
        file_name : str
        sicd_meta : SICDType
        user_data : None|Dict[str, str]
        check_older_version : bool
            Try to use an older version (1.1) of the SICD standard, for possible
            application compliance issues?
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        if check_existence and os.path.exists(file_name):
            raise SarpyIOError('Given file {} already exists, and a new SIO file cannot be created here.'.format(file_name))

        # choose magic number (with user data) and corresponding endian-ness
        magic_number = 0xFD7F02FF
        endian = SIODetails.ENDIAN[magic_number]

        # define basic image details
        image_size = (sicd_meta.ImageData.NumRows, sicd_meta.ImageData.NumCols)
        pixel_type = sicd_meta.ImageData.PixelType
        if pixel_type == 'RE32F_IM32F':
            data_type = numpy.dtype('{}f4'.format(endian))
            element_type = 13
            element_size = 8
            transform_data = 'COMPLEX'
        elif pixel_type == 'RE16I_IM16I':
            data_type = numpy.dtype('{}i2'.format(endian))
            element_type = 12
            element_size = 4
            transform_data = complex_to_int
        else:
            data_type = numpy.dtype('{}u1'.format(endian))
            element_type = 11
            element_size = 2
            transform_data = complex_to_amp_phase(sicd_meta.ImageData.AmpTable)
        # construct the sio header
        header = numpy.array(
            [magic_number, image_size[0], image_size[1], element_type, element_size],
            dtype='>u4')
        # construct the user data - must be {str : str}
        if user_data is None:
            user_data = {}
        uh_args = sicd_meta.get_des_details(check_older_version)
        user_data['SICDMETA'] = sicd_meta.to_xml_string(tag='SICD', urn=uh_args['DESSHTN'])
        data_offset = 20
        with open(file_name, 'wb') as fi:
            fi.write(struct.pack('{}5I'.format(endian), *header))
            # write the user data - name size, name, value size, value
            for name in user_data:
                name_bytes = name.encode('utf-8')
                fi.write(struct.pack('{}I'.format(endian), len(name_bytes)))
                fi.write(struct.pack('{}{}s'.format(endian, len(name_bytes)), name_bytes))
                val_bytes = user_data[name].encode('utf-8')
                fi.write(struct.pack('{}I'.format(endian), len(val_bytes)))
                fi.write(struct.pack('{}{}s'.format(endian, len(val_bytes)), val_bytes))
                data_offset += 4 + len(name_bytes) + 4 + len(val_bytes)
        # initialize the bip writer - we're ready to go
        output_bands = 2
        super(SIOWriter, self).__init__(file_name, image_size, data_type, output_bands,
                                        transform_data=transform_data, data_offset=data_offset)



"""
Module for reading and writing SIDD files - should support SIDD version 1.0 and above.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import sys
from functools import reduce
import re
from typing import List, Union, BinaryIO
from datetime import datetime

import numpy

from sarpy.io.xml.base import parse_xml_from_string
from sarpy.io.general.utils import is_file_like
from sarpy.io.general.base import AggregateChipper, SarpyIOError
from sarpy.io.general.nitf import NITFDetails, NITFReader, NITFWriter, ImageDetails, DESDetails, \
    image_segmentation, get_npp_block, interpolate_corner_points_string
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, XMLDESSubheader
from sarpy.io.general.nitf_elements.security import NITFSecurityTags
from sarpy.io.general.nitf_elements.image import ImageSegmentHeader, ImageBands, ImageBand

from sarpy.io.product.base import SIDDTypeReader
from sarpy.io.product.sidd2_elements.SIDD import SIDDType
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd import extract_clas as extract_clas_sicd


logger = logging.getLogger(__name__)

########
# module variables
_class_priority = {'U': 0, 'R': 1, 'C': 2, 'S': 3, 'T': 4}


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name):
    """
    Tests whether a given file_name corresponds to a SIDD file.
    Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    SIDDReader|None
        `SIDDReader` instance if SIDD file, `None` otherwise
    """

    try:
        nitf_details = SIDDDetails(file_name)
        if nitf_details.is_sidd:
            logger.info('File {} is determined to be a SIDD (NITF format) file.'.format(file_name))
            return SIDDReader(nitf_details)
        else:
            return None
    except SarpyIOError:
        # we don't want to catch parsing errors, for now
        return None


#########
# Helper object for initially parses NITF header - specifically looking for SICD elements


class SIDDDetails(NITFDetails):
    """
    SIDD are stored in NITF 2.1 files.
    """

    __slots__ = (
        '_is_sidd', '_sidd_meta', '_sicd_meta',
        'img_segment_rows', 'img_segment_columns')

    def __init__(self, file_object):
        """

        Parameters
        ----------
        file_object : str|BinaryIO
            file name or file like object for a NITF 2.1 or 2.0 containing a SIDD
        """

        self._img_headers = None
        self._is_sidd = False
        self._sidd_meta = None
        self._sicd_meta = None
        super(SIDDDetails, self).__init__(file_object)
        if self._nitf_header.ImageSegments.subhead_sizes.size == 0:
            raise SarpyIOError('There are no image segments defined.')
        if self._nitf_header.GraphicsSegments.item_sizes.size > 0:
            raise SarpyIOError('A SIDD file does not allow for graphics segments.')
        if self._nitf_header.DataExtensions.subhead_sizes.size == 0:
            raise SarpyIOError(
                'A SIDD file requires at least one data extension, containing the '
                'SIDD xml structure.')
        # define the sidd and sicd metadata
        self._find_sidd()
        if not self.is_sidd:
            raise SarpyIOError('Could not find SIDD xml data extensions.')
        # populate the image details
        self.img_segment_rows = numpy.zeros(self.img_segment_offsets.shape, dtype=numpy.int64)
        self.img_segment_columns = numpy.zeros(self.img_segment_offsets.shape, dtype=numpy.int64)
        for i, im_header in enumerate(self.img_headers):
            self.img_segment_rows[i] = im_header.NROWS
            self.img_segment_columns[i] = im_header.NCOLS

    @property
    def is_sidd(self):
        """
        bool: whether file name corresponds to a SIDD file, or not.
        """

        return self._is_sidd

    @property
    def sidd_meta(self):
        """
        None|List[sarpy.io.product.sidd2_elements.SIDD.SIDDType]: the sidd meta-data structure(s).
        """

        return self._sidd_meta

    @property
    def sicd_meta(self):
        """
        None|List[sarpy.io.complex.sicd_elements.SICD.SICDType]: the sicd meta-data structure(s).
        """

        return self._sicd_meta

    def _find_sidd(self):
        self._is_sidd = False
        if self.des_subheader_offsets is None:
            return

        self._sidd_meta = []
        self._sicd_meta = []

        for i in range(self.des_subheader_offsets.size):
            subhead_bytes = self.get_des_subheader_bytes(i)
            if subhead_bytes.startswith(b'DEXML_DATA_CONTENT'):
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:
                        self._is_sidd = True
                        self._sidd_meta.append(SIDDType.from_node(root_node, xml_ns, ns_key='default'))
                    elif 'SICD' in root_node.tag:
                        self._sicd_meta.append(SICDType.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error('Failed checking des xml header at index {} with error {}'.format(i, e))
                    continue
            elif subhead_bytes.startswith(b'DESIDD_XML'):
                # This is an old format SIDD header
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SIDD' in root_node.tag:
                        self._is_sidd = True
                        self._sidd_meta.append(SIDDType.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error(
                        'We found an apparent old-style SIDD DES header at index {},\n\t'
                        'but failed parsing with error {}'.format(i, e))
                    continue
            elif subhead_bytes.startswith(b'DESICD_XML'):
                # This is an old format SICD header
                des_bytes = self.get_des_bytes(i)
                try:
                    root_node, xml_ns = parse_xml_from_string(des_bytes)
                    if 'SICD' in root_node.tag:
                        self._sicd_meta.append(SICDType.from_node(root_node, xml_ns, ns_key='default'))
                except Exception as e:
                    logger.error(
                        'We found an apparent old-style SICD DES header at index {},\n\t'
                        'but failed parsing with error {}'.format(i, e))
                    continue

        if not self._is_sidd:
            return

        for sicd in self._sicd_meta:
            sicd.derive()


#######
#  The actual reading implementation

def _check_iid_format(iid1, i):
    if not (iid1[:4] == 'SIDD' and iid1[4:].isnumeric()):
        raise ValueError('Got poorly formatted image segment id {} at position {}'.format(iid1, i))


class SIDDReader(NITFReader, SIDDTypeReader):
    """
    A reader object for a SIDD file (NITF container with SICD contents)
    """

    def __init__(self, nitf_details):
        """

        Parameters
        ----------
        nitf_details : str|BinaryIO|SIDDDetails
            filename, file-like object, or SIDDDetails object
        """

        if isinstance(nitf_details, str) or is_file_like(nitf_details):
            nitf_details = SIDDDetails(nitf_details)
        if not isinstance(nitf_details, SIDDDetails):
            raise TypeError('The input argument for SIDDReader must be a filename or '
                            'SIDDDetails object.')

        if not nitf_details.is_sidd:
            raise ValueError(
                'The input file passed in appears to be a NITF 2.1 file that does not contain '
                'valid sidd metadata.')

        self._nitf_details = nitf_details
        SIDDTypeReader.__init__(self, self.nitf_details.sidd_meta, self.nitf_details.sicd_meta)
        NITFReader.__init__(self, nitf_details, reader_type="SIDD")

    @property
    def nitf_details(self):
        # type: () -> SIDDDetails
        """
        SIDDDetails: The SIDD NITF details object.
        """

        return self._nitf_details

    def _find_segments(self):
        # determine image segmentation from image headers
        segments = [[] for _ in self._sidd_meta]
        for i, img_header in enumerate(self.nitf_details.img_headers):
            # skip anything but SAR for now (i.e. legend)
            if img_header.ICAT != 'SAR':
                continue

            iid1 = img_header.IID1  # required to be of the form SIDD######
            _check_iid_format(iid1, i)
            element = int(iid1[4:7])
            if element > len(self._sidd_meta):
                raise ValueError('Got image segment id {}, but there are only {} '
                                 'sidd elements'.format(iid1, len(self._sidd_meta)))
            segments[element - 1].append(i)
        # Ensure that all segments are populated
        for i, entry in enumerate(segments):
            if len(entry) < 1:
                raise ValueError('Did not find any image segments for SIDD {}'.format(i))
        return segments

    def _check_img_details(self, segment):
        raw_dtype, output_dtype, raw_bands, output_bands, transform_data = self._extract_chipper_params(segment[0])
        for this_index in segment[1:]:
            this_raw_dtype, this_output_dtype, this_raw_bands, this_output_bands, \
                _ = self._extract_chipper_params(this_index)
            if this_raw_dtype.name != raw_dtype.name:
                raise ValueError(
                    'Segments at index {} and {} have incompatible data types '
                    '{} and {}'.format(segment[0], this_index, raw_dtype, this_raw_dtype))
            if this_output_dtype.name != output_dtype.name:
                raise ValueError(
                    'Segments at index {} and {} have incompatible output data types '
                    '{} and {}'.format(segment[0], this_index, output_dtype, this_output_dtype))
            if this_raw_bands != raw_bands:
                raise ValueError(
                    'Segments at index {} and {} have incompatible input bands '
                    '{} and {}'.format(segment[0], this_index, raw_bands, this_raw_bands))
            if this_output_bands != output_bands:
                raise ValueError(
                    'Segments at index {} and {} have incompatible output bands '
                    '{} and {}'.format(segment[0], this_index, output_bands, this_output_bands))
        return raw_dtype, output_dtype, raw_bands, output_bands, transform_data

    def _construct_chipper(self, segment, index):
        # get the image size
        sidd = self.sidd_meta[index]
        rows = sidd.Measurement.PixelFootprint.Row
        cols = sidd.Measurement.PixelFootprint.Col

        # extract the basic elements for the chippers
        raw_dtype, output_dtype, raw_bands, output_bands, transform_data = self._check_img_details(segment)
        if len(segment) == 1:
            return self._define_chipper(
                segment[0], raw_dtype=raw_dtype, raw_bands=raw_bands, transform_data=transform_data,
                output_dtype=output_dtype, output_bands=output_bands)
        else:
            # get the bounds definition
            bounds = self._get_chipper_partitioning(segment, rows, cols)
            # define the chippers
            chippers = [
                self._define_chipper(img_index, raw_dtype=raw_dtype, raw_bands=raw_bands, transform_data=transform_data,
                                     output_dtype=output_dtype, output_bands=output_bands) for img_index in segment]
            # define the aggregate chipper
            return AggregateChipper(bounds, output_dtype, chippers, output_bands=output_bands)


#########
# The writer implementation

def validate_sidd_for_writing(sidd_meta):
    """
    Helper method which ensures the provided SIDD structure is appropriate.

    Parameters
    ----------
    sidd_meta : SIDDType|List[SIDDType]|SIDDType1|List[SIDDType1]

    Returns
    -------
    Tuple[Union[SIDDType, SIDDType1]]
    """

    def inspect_sidd(the_sidd):
        # we must have the image size
        if the_sidd.Measurement is None:
            raise ValueError('The sidd_meta has un-populated Measurement, '
                             'and nothing useful can be inferred.')
        if the_sidd.Measurement.PixelFootprint is None:
            raise ValueError('The sidd_meta has un-populated Measurement.PixelFootprint, '
                             'and this is not valid for writing.')

        # we must have the pixel type
        if the_sidd.Display is None:
            raise ValueError('The sidd_meta has un-populated Display, '
                             'and nothing useful can be inferred.')
        if the_sidd.Display.PixelType is None:
            raise ValueError('The sidd_meta has un-populated Display.PixelType, '
                             'and nothing useful can be inferred.')
        # No support for LUT until necessary
        if the_sidd.Display.PixelType in ('MONO8LU', 'RGB8LU'):
            raise ValueError('PixelType requiring lookup table currently unsupported.')

        # we must have collection time
        if the_sidd.ExploitationFeatures is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.')
        if the_sidd.ExploitationFeatures.Collections is None or len(the_sidd.ExploitationFeatures.Collections) == 0:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.')
        if the_sidd.ExploitationFeatures.Collections[0].Information is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.Information.')
        if the_sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime is None:
            raise ValueError('The sidd_meta has un-populated ExploitationFeatures.Collections.'
                             'Information.CollectionDateTime')

        # try validating versus the appropriate schema
        xml_str = the_sidd.to_xml_string(tag='SIDD')
        from sarpy.consistency.sidd_consistency import evaluate_xml_versus_schema
        if isinstance(the_sidd, SIDDType1):
            urn = 'urn:SIDD:1.0.0'
        elif isinstance(the_sidd, SIDDType):
            urn = 'urn:SIDD:2.0.0'
        else:
            raise ValueError('Unhandled type {}'.format(type(the_sidd)))
        result = evaluate_xml_versus_schema(xml_str, urn)
        if result is False:
            logger.warning(
                'The provided SIDD does not properly validate\n\t'
                'against the schema for {}'.format(urn))

    if isinstance(sidd_meta, (SIDDType, SIDDType1)):
        inspect_sidd(sidd_meta)
        return (sidd_meta, )
    elif isinstance(sidd_meta, (tuple, list)):
        out = []
        for entry in sidd_meta:
            if not isinstance(entry, (SIDDType, SIDDType1)):
                raise TypeError('All entries are required to be an instance of SIDDType, '
                                'got type {}'.format(type(entry)))
            inspect_sidd(entry)
            out.append(entry)
        return tuple(out)
    else:
        raise TypeError('sidd_meta is required to be an instance of SIDDType or a list/tuple '
                        'of such instances, got {}'.format(type(sidd_meta)))


def validate_sicd_for_writing(sicd_meta):
    """
    Helper method which ensures the provided SICD structure is appropriate.

    Parameters
    ----------
    sicd_meta : SICDType|List[SICDType]

    Returns
    -------
    None|Tuple[SICDType]
    """

    if sicd_meta is None:
        return None
    if isinstance(sicd_meta, SICDType):
        return (sicd_meta, )
    elif isinstance(sicd_meta, (tuple, list)):
        out = []
        for entry in sicd_meta:
            if not isinstance(entry, SICDType):
                raise TypeError('All entries are required to be an instance of SICDType, '
                                'got type {}'.format(type(entry)))
            out.append(entry)
        return tuple(out)
    else:
        raise TypeError('sicd_meta is required to be an instance of SICDType or a list/tuple '
                        'of such instances, got {}'.format(type(sicd_meta)))


def extract_clas(the_sidd):
    """
    Extract the classification string from a SIDD as appropriate for NITF Security
    tags CLAS attribute.

    Parameters
    ----------
    the_sidd : SIDDType|SIDDType1

    Returns
    -------
    str
    """

    class_str = the_sidd.ProductCreation.Classification.classification

    if class_str is None or class_str == '':
        return 'U'
    else:
        return class_str[:1]


def extract_clsy(the_sidd):
    """
    Extract the ownerProducer string from a SIDD as appropriate for NITF Security
    tags CLSY attribute.

    Parameters
    ----------
    the_sidd : SIDDType|SIDDType1

    Returns
    -------
    str
    """

    owner = the_sidd.ProductCreation.Classification.ownerProducer.upper()
    if owner is None:
        return ''
    elif owner in ('USA', 'CAN', 'AUS', 'NZL'):
        return owner[:2]
    elif owner == 'GBR':
        return 'UK'
    elif owner == 'NATO':
        return 'XN'
    else:
        logger.warning(
            'Got owner {}, and the CLSY will be truncated\n\t'
            'to two characters.'.format(owner))
        return owner[:2]


class SIDDWriter(NITFWriter):
    """
    Writer class for a SIDD file - a NITF file containing data derived from complex radar data and
    SIDD and SICD data extension(s).
    """

    __slots__ = ('_sidd_meta', '_sicd_meta', )

    def __init__(self, file_name, sidd_meta, sicd_meta, check_existence=True):
        """

        Parameters
        ----------
        file_name : str
        sidd_meta : SIDDType|List[SIDDType]|SIDDType1|List[SIDDType1]
        sicd_meta : SICDType|List[SICDType]
        check_existence : bool
            Should we check if the given file already exists, and raises an exception if so?
        """

        self._shapes = None
        self._sidd_meta = validate_sidd_for_writing(sidd_meta)
        self._set_shapes()
        self._sicd_meta = validate_sicd_for_writing(sicd_meta)
        self._security_tags = None
        self._nitf_header = None
        self._img_groups = None
        self._img_details = None
        self._des_details = None
        super(SIDDWriter, self).__init__(file_name, check_existence=check_existence)

    @property
    def sidd_meta(self):
        """
        Tuple[SIDDType]: The sidd metadata.
        """

        return self._sidd_meta

    @property
    def sicd_meta(self):
        """
        None|Tuple[SICDType]: The sicd metadata.
        """

        return self._sicd_meta

    def _set_shapes(self):
        if isinstance(self._sidd_meta, tuple):
            self._shapes = tuple(
                (entry.Measurement.PixelFootprint.Row, entry.Measurement.PixelFootprint.Col)
                for entry in self._sidd_meta)
        else:
            self._shapes = (
                (self._sidd_meta.Measurement.PixelFootprint.Row,
                 self._sidd_meta.Measurement.PixelFootprint.Col), )

    def _get_security_tags(self, index):
        """
        Gets the security tags for SIDD at `index`.

        Parameters
        ----------
        index : int

        Returns
        -------
        NITFSecurityTags
        """

        args = {}
        sidd = self.sidd_meta[index]
        if sidd.ProductCreation is not None and sidd.ProductCreation.Classification is not None:
            args['CLAS'] = extract_clas(sidd)
            args['CLSY'] = extract_clsy(sidd)

        security = sidd.NITF.get('Security', {})
        # noinspection PyProtectedMember
        for key in NITFSecurityTags._ordering:
            if key in security:
                args[key] = security[key]
        return NITFSecurityTags(**args)

    def _create_security_tags(self):
        def class_priority(cls1, cls2):
            p1 = _class_priority[cls1]
            p2 = _class_priority[cls2]
            if p1 >= p2:
                return cls1
            return cls2

        # determine the highest priority clas string from all the sidds & sicds
        clas_collection = [extract_clas(sidd) for sidd in self.sidd_meta]
        if self.sicd_meta is not None:
            clas_collection.extend([extract_clas_sicd(sicd) for sicd in self.sicd_meta])
        clas = reduce(class_priority, clas_collection)
        # try to determine clsy from all sidds
        clsy_collection = list(set(extract_clsy(sidd) for sidd in self.sidd_meta))
        clsy = clsy_collection[0] if len(clsy_collection) else None
        # populate the attribute
        self._security_tags = NITFSecurityTags(CLAS=clas, CLSY=clsy)

    def _get_iid2(self, index):
        """
        Get the IID2 for the sidd at `index`.

        Parameters
        ----------
        index : int

        Returns
        -------
        str
        """

        sidd = self.sidd_meta[index]

        iid2 = sidd.NITF.get('IID2', None)
        if iid2 is None:
            iid2 = sidd.NITF.get('FTITLE', None)
        if iid2 is None:
            iid2 = sidd.NITF.get('SUGGESTED_NAME', None)
        if iid2 is None:
            iid2 = 'SIDD: Unknown'
        return iid2

    def _get_ftitle(self, index=0):
        sidd = self.sidd_meta[index]
        ftitle = sidd.NITF.get('FTITLE', None)
        if ftitle is None:
            ftitle = sidd.NITF.get('SUGGESTED_NAME', None)
        if ftitle is None:
            ftitle = 'SIDD: Unknown'
        return ftitle

    def _get_fdt(self):
        sidd = self.sidd_meta[0]
        if sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime is not None:
            the_time = sidd.ExploitationFeatures.Collections[0].Information.CollectionDateTime.astype('datetime64[s]')
            return re.sub(r'[^0-9]', '', str(the_time))
        else:
            return super(SIDDWriter, self)._get_fdt()

    def _get_ostaid(self, index=0):
        sidd = self.sidd_meta[index]
        ostaid = sidd.NITF.get('OSTAID', 'Unknown')
        return ostaid

    def _get_isorce(self, index=0):
        sidd = self.sidd_meta[index]
        isorce = sidd.NITF.get('ISORCE', sidd.ExploitationFeatures.Collections[0].Information.SensorName)
        if isorce is None:
            isorce = 'Unknown'
        return isorce

    def _image_parameters(self, index):
        """
        Get the image parameters for the sidd at `index`.

        Parameters
        ----------
        index

        Returns
        -------
        (int, int, str, numpy.dtype, Union[bool, str, callable], str, int, tuple, tuple)
            pixel_size - the size of each pixel in bytes.
            abpp - the actual bits per pixel.
            irep - the image representation
            raw_dtype - the data type.
            transform_data -
            pv_type - the pixel type string.
            band_count - the number of bands
            irepband - the image representation.
            im_segments - Segmentation of the form `((row start, row end, column start, column end))`
        """

        sidd = self.sidd_meta[index]
        assert isinstance(sidd, (SIDDType, SIDDType1))
        if sidd.Display.PixelType == 'MONO8I':
            pixel_size = 1
            abpp = 8
            irep = 'MONO'
            raw_dtype = numpy.dtype('>u1')
            transform_data = None
            pv_type = 'INT'
            band_count = 1
            irepband = ('M', )
        elif sidd.Display.PixelType == 'MONO16I':
            pixel_size = 2
            abpp = 16
            irep = 'MONO'
            raw_dtype = numpy.dtype('>u2')
            transform_data = None
            pv_type = 'INT'
            band_count = 1
            irepband = ('M', )
        elif sidd.Display.PixelType == 'RGB24I':
            pixel_size = 3
            abpp = 8
            irep = 'RGB'
            raw_dtype = numpy.dtype('>u1')
            transform_data = None
            pv_type = 'INT'
            band_count = 3
            irepband = ('R', 'G', 'B')
        else:
            raise ValueError('Unsupported PixelType {}'.format(sidd.Display.PixelType))

        image_segment_limits = image_segmentation(
            sidd.Measurement.PixelFootprint.Row, sidd.Measurement.PixelFootprint.Col, pixel_size)
        return pixel_size, abpp, irep, raw_dtype, transform_data, pv_type, band_count, irepband, image_segment_limits

    @staticmethod
    def _get_icp(sidd):
        """
        Get the Image corner point array, if possible.

        Parameters
        ----------
        sidd : SIDDType|SIDDType1

        Returns
        -------
        numpy.ndarray
        """

        if isinstance(sidd, SIDDType) and sidd.GeoData is not None and sidd.GeoData.ImageCorners is not None:
            return sidd.GeoData.ImageCorners.get_array(dtype=numpy.dtype('float64'))
        elif isinstance(sidd, SIDDType1) and sidd.GeographicAndTarget is not None and \
                sidd.GeographicAndTarget.GeographicCoverage is not None and \
                sidd.GeographicAndTarget.GeographicCoverage.Footprint is not None:
            return sidd.GeographicAndTarget.GeographicCoverage.Footprint.get_array(dtype=numpy.dtype('float64'))
        return None

    def _create_image_segment(self, index, img_groups, img_details):
        cur_count = len(img_details)
        pixel_size, abpp, irep, raw_dtype, transform_data, pv_type, band_count, irepband, \
            image_segment_limits = self._image_parameters(index)
        new_count = len(image_segment_limits)
        img_groups.append(tuple(range(cur_count, cur_count+new_count)))
        security = self._get_security_tags(index)

        iid2 = self._get_iid2(index)
        sidd = self.sidd_meta[index]

        idatim = self._get_fdt()

        isorce = self._get_isorce(index)

        rows = sidd.Measurement.PixelFootprint.Row
        cols = sidd.Measurement.PixelFootprint.Col

        icp = self._get_icp(sidd)
        bands = [ImageBand(ISUBCAT='', IREPBAND=entry) for entry in irepband]

        for i, entry in enumerate(image_segment_limits):
            this_rows = entry[1]-entry[0]
            this_cols = entry[3]-entry[2]
            subhead = ImageSegmentHeader(
                IID1='SIDD{0:03d}{1:03d}'.format(len(img_groups), i+1),
                IDATIM=idatim,
                IID2=iid2,
                ISORCE=isorce,
                IREP=irep,
                ICAT='SAR',
                NROWS=this_rows,
                NCOLS=this_cols,
                PVTYPE=pv_type,
                ABPP=abpp,
                IGEOLO=interpolate_corner_points_string(numpy.array(entry, dtype=numpy.int64), rows, cols, icp),
                IMODE='P',
                NPPBH=get_npp_block(this_cols),
                NPPBV=get_npp_block(this_rows),
                NBPP=abpp,
                IDLVL=i+1+len(img_details),
                IALVL=i+len(img_details),
                ILOC='{0:05d}{1:05d}'.format(entry[0], entry[2]),
                Bands=ImageBands(values=bands),
                Security=security)
            img_details.append(ImageDetails(band_count, raw_dtype, transform_data, entry, subhead))

    def _create_image_segment_details(self):
        super(SIDDWriter, self)._create_image_segment_details()
        img_groups = []
        img_details = []
        for index in range(len(self.sidd_meta)):
            self._create_image_segment(index, img_groups, img_details)
        self._img_groups = tuple(img_groups)
        self._img_details = tuple(img_details)

    def _create_sidd_des_segment(self, index):
        """
        Create the details for the sidd data extension at `index`.

        Parameters
        ----------
        index : int

        Returns
        -------
        DESDetails
        """

        imgs = self._img_groups[index]
        security = self.image_details[imgs[0]].subheader.Security
        sidd = self.sidd_meta[index]
        uh_args = sidd.get_des_details()

        try:
            desshdt = str(sidd.ProductCreation.ProcessorInformation.ProcessingDateTime.astype('datetime64[s]'))
        except AttributeError:
            desshdt = str(numpy.datetime64('now'))
        if desshdt[-1] != 'Z':
            desshdt += 'Z'
        uh_args['DESSHDT'] = desshdt

        desshlpg = ''
        icp = self._get_icp(sidd)
        if icp is not None:
            temp = []
            for entry in icp:
                temp.append('{0:0=+12.8f}{1:0=+13.8f}'.format(entry[0], entry[1]))
            temp.append(temp[0])
            desshlpg = ''.join(temp)
        uh_args['DESSHLPG'] = desshlpg
        subhead = DataExtensionHeader(
            Security=security,
            UserHeader=XMLDESSubheader(**uh_args))
        return DESDetails(subhead, sidd.to_xml_bytes(tag='SIDD'))

    def _create_sicd_des_segment(self, index):
        """
        Create the details for the sicd data extension at `index`.

        Parameters
        ----------
        index : int

        Returns
        -------
        DESDetails
        """

        security_tags = self.security_tags
        sicd = self.sicd_meta[index]
        uh_args = sicd.get_des_details(check_version1_compliance=True)
        if sicd.ImageCreation.DateTime is None:
            desshdt = datetime.utcnow().isoformat('T', timespec='seconds')
        else:
            desshdt = str(sicd.ImageCreation.DateTime.astype('datetime64[s]'))
        if desshdt[-1] != 'Z':
            desshdt += 'Z'
        uh_args['DESSHDT'] = desshdt

        desshlpg = ''
        if sicd.GeoData is not None and sicd.GeoData.ImageCorners is not None:
            # noinspection PyTypeChecker
            icp = sicd.GeoData.ImageCorners.get_array(dtype=numpy.float64)
            temp = []
            for entry in icp:
                temp.append('{0:0=+12.8f}{1:0=+13.8f}'.format(entry[0], entry[1]))
            temp.append(temp[0])
            desshlpg = ''.join(temp)
        uh_args['DESSHLPG'] = desshlpg

        subhead = DataExtensionHeader(
            Security=security_tags,
            UserHeader=XMLDESSubheader(**uh_args))

        return DESDetails(subhead, sicd.to_xml_bytes(tag='SICD', urn=uh_args['DESSHTN']))

    def _create_data_extension_details(self):
        super(SIDDWriter, self)._create_data_extension_details()
        des_details = []
        for index in range(len(self.sidd_meta)):
            des_details.append(self._create_sidd_des_segment(index))
        if self.sicd_meta is not None:
            for index in range(len(self.sicd_meta)):
                des_details.append(self._create_sicd_des_segment(index))
        self._des_details = tuple(des_details)



"""
Basic First Example
===================

This indicates the basic functionality for reading a complex format data set. It
is intended that this script will be read, and the relevant portions run
individually.


Learning Python
----------------

It seems somewhat common for some users of sarpy to be experts in SAR, but not
well versed with Python. If this describes you (and even if it doesn't), the
built-in `help()` function can often help clear up confusion found during an
interactive Python session.

Check out a given module, class, or function using it's path
>>> help('sarpy')
or
>>> help('sarpy.io.complex.sicd.SICDReader')

or, check out a variable that has been defined somehow
>>> from sarpy.io.complex.converter import open_complex
>>> reader = open_complex('<path to file>')
>>> help(reader)
"""


"""
General file opening
--------------------

Open a **complex format file** using a general purpose opener from a specified 
file name.

This general purpose opener iterates over the reader objects defined in
sarpy.io.complex, trying each and attempting to catch exceptions. It returns
the first one that works, or raises an exception if none of the readers work.
Note that this exception catching process, though we attempt to make everything
completely robust, can sometimes confusingly hide errors in the file identification,
parsing and/or interpretation.

It is important to note that this opener will not open any file that is not
**complex format** (like a WBID or detected image).
"""
from sarpy.io.complex.converter import open_complex


reader = open_complex('<path to file>')
# this will return an instance of one of the reader classes defined in the modules in sarpy.io.complex
print(type(reader))


"""
Direct file opening
-------------------

If you know the file type of your file, you may want to use the correct reader
class directly, especially if the general opener is not working.
"""
from sarpy.io.complex.sicd import SICDReader
reader = SICDReader('<path to file>')  # same class as referenced above


"""
SICD metadata structure
-----------------------

Access the SICD structure or tuple of structures associated with the reader.
Note that the sicd structure is defined using elements of sarpy.io.complex.sicd_elements

A sicd file will necessarily be composed of a single image, but other file formats
(like sentinel or radarsat) often contain multiple images combined into a single
package (i.e. multiple polarizations or other aggregate collections).
"""

# nebulous contents - this will be an instance of the sicd structure, or a tuple
# of sicd structures
nebulous_contents = reader.sicd_meta

# Unified access - this will always be a tuple of sicd structures,
# with one sicd structure per image
sicd_tuple = reader.get_sicds_as_tuple()

the_sicd = sicd_tuple[0]  # access the desired sicd structure
# provide a human readable, if long, contents to terminal
print(the_sicd)
# get xml string representation
xml_string = the_sicd.to_xml_string(tag='SICD')
# get json friendly dict representation
dict_representation = the_sicd.to_dict()

# access field values
print(the_sicd.CollectionInfo.CollectorName)


"""
Read complex pixel data
-----------------------

The recommended methodology uses slice notation. The basic syntax is as:
>>> data = reader[row_start:row_end:row_step, col_start:col_end:col_step, image_index=0]
"""
# in the event of a single image segment
all_data = reader[:]  # or reader[:, :] - reads all data
decimated_data = reader[::10, ::10] # reads every 10th pixel

# in the event of multiple image segments, use another slice index like
all_data = reader[:, :, 0]

"""
Opening remote file
-------------------

The SICD reader (and also SIDD reader) have been implemented to accept binary file-like 
objects, specifically intended to enable 

**Speed/efficiency:** Files read using the file system (i.e. via file name
or local file-like object) are read efficiently via numpy memory map. Reading
across a network file system, commonly encountered as reading from a file-share
drive mounted to your local system, maintains the efficiency of numpy memory map
usage, but the speed will be impacted (perhaps significantly) by network latency.

It should be noted that the flexibility of reading using a file-like object comes
at a significant efficiency and speed cost, particularly for reading decimated or
down-selected data. A numpy memory map can not be utilized (at least as of May 2021)
for a non-file object, and reading/interpreting data becomes a fully manual
and non-optimized process. The entire continguous chunk of data containing the
desired segment of data will be read, then down-selected. This is to accommodate
for the overhead of the connection request for remote reading - simple bench marks
indicate that the bottleneck for performing a remote read is clearly the connection
request, and presents no good opportunity for clear optimization.
"""

# for the purposes of general purpose example, we reference a basic example sicd
# file hosted for the SIX project usage. It is recommended to use local files, as
# described below
import smart_open
file_object = smart_open.open(
    'https://six-library.s3.amazonaws.com/sicd_example_RMA_RGZERO_RE32F_IM32F_cropped_multiple_image_segments.nitf',
    mode='rb',  # must be opened in binary mode
    buffering=4*1024*1024)  # it has been observed that setting a manual buffer size may help



"""
Basic data plot
---------------

Show some basic plots of the data.
 
**Note:** the sarpy_apps project provides robust interactive tools.
"""

from matplotlib import pyplot
from sarpy.visualization.remap import Density

remap_function = Density()

fig, axs = pyplot.subplots(nrows=1, ncols=1, figsize=(5, 5))
axs.imshow(remap_function(all_data), cmap='gray')
pyplot.show()


"""
Convert to SICD format
----------------------

Convert a complex dataset, in any format handled by sarpy, to SICD
"""

from sarpy.io.complex.converter import conversion_utility
conversion_utility('<complex format file>', '<output_directory>')
